import time


class CacheMixin:

    @property
    def _cache(self):
        cache_dict = getattr(self, '_cache_dict', None)
        if cache_dict is None:
            cache_dict = dict()
            self._cache_dict = cache_dict
        return cache_dict

    def cacheget(self, key):
        value, expires = self._cache.get(key, (None, None))

        if value is not None and expires is not None and time.time() > expires:
            del self._cache[key]
            value = None

        return value

    def cacheset(self, key, value, seconds=None):
        expires = time.time() + seconds if seconds else None
        self._cache[key] = (value, expires)
        return (value, expires)



import datetime


def iso_8601_duration(seconds):
    hours = seconds // 3600
    seconds = seconds % 3600
    minutes = seconds // 60
    seconds = seconds % 60

    formatted_date = "PT"
    if hours:
        formatted_date = "%s%dH" % (formatted_date, hours)
    if minutes:
        formatted_date = "%s%dM" % (formatted_date, minutes)
    if seconds or formatted_date == "PT":
        formatted_date = "%s%dS" % (formatted_date, seconds)

    return formatted_date


def timestamp(utc_dt=None):
    """
    Returns the utc timestamp of now or the given utc_dt object.
    """
    epoch = datetime.datetime.utcfromtimestamp(0)
    dt = utc_dt or datetime.datetime.utcnow()
    return (dt - epoch).total_seconds()



import datetime
import unittest

from zaach.time import formatter


class FormatterTests(unittest.TestCase):

    def test_iso_8601_duration_1(self):
        self.assertEqual(
            formatter.iso_8601_duration(0),
            "PT0S"
        )

    def test_iso_8601_duration_2(self):
        self.assertEqual(
            formatter.iso_8601_duration(1),
            "PT1S"
        )

    def test_iso_8601_duration_3(self):
        self.assertEqual(
            formatter.iso_8601_duration(60),
            "PT1M"
        )

    def test_iso_8601_duration_4(self):
        self.assertEqual(
            formatter.iso_8601_duration(70),
            "PT1M10S"
        )

    def test_iso_8601_duration_5(self):
        self.assertEqual(
            formatter.iso_8601_duration(3600),
            "PT1H"
        )

    def test_iso_8601_duration_6(self):
        self.assertEqual(
            formatter.iso_8601_duration(3610),
            "PT1H10S"
        )

    def test_iso_8601_duration_7(self):
        self.assertEqual(
            formatter.iso_8601_duration(3660),
            "PT1H1M"
        )

    def test_iso_8601_duration_8(self):
        self.assertEqual(
            formatter.iso_8601_duration(3661),
            "PT1H1M1S"
        )

    def test_timestamp_1(self):
        ts = formatter.timestamp()
        self.assertTrue(ts)

    def test_timestamp_2(self):
        ts = formatter.timestamp(datetime.datetime(1970, 1, 1))
        self.assertEqual(ts, 0)

    def test_timestamp_3(self):
        if hasattr(datetime.datetime, 'timestamp'):  # python3
            ts1 = formatter.timestamp()
            ts2 = datetime.datetime.timestamp(datetime.datetime.utcnow())
            self.assertTrue((ts2 - ts1) < 1)


if __name__ == '__main__':
    unittest.main()



import datetime


CET_OFFSET = 1
CEST_OFFSET = 2


def get_last_sunday(year, month):
    """
    Returns the last sunday in a given year of a given month as a date.
    """
    for day in range(31, 0, -1):
        try:
            dt = datetime.date(year, month, day)
        except ValueError:
            continue
        else:
            weekday = dt.weekday()
            if weekday != 6:
                dt = dt - datetime.timedelta(days=weekday + 1)
            return dt


def get_utc_offset_reference(utc_dt):
    """
    Returns the offset from UTC to Central European (Summer) time, in hours.
    This implementation is not very performant and should only be used for
    edge cases and as a reference implementation to test against.
    """
    year = utc_dt.year

    sun = get_last_sunday(year, 3)
    cest_barrier = datetime.datetime(year, 3, sun.day, 1)

    sun = get_last_sunday(year, 10)
    cet_barrier = datetime.datetime(year, 10, sun.day, 1)

    if utc_dt < cest_barrier:  # 1.1. till xx.03.
        return CET_OFFSET
    elif utc_dt < cet_barrier:  # xx.3. till xx.10.
        return CEST_OFFSET
    else:  # xx.10. till 31.12.
        return CET_OFFSET


def get_utc_offset(utc_dt):
    """
    Returns the offset from UTC to Central European (Summer) time, in hours.
    This implementation is way more performant, due to the following facts:
    * November, December, January, February are CET
    * March is CET up to the last sunday, 1am UTC, then CEST
    * April, May, June, July, August, September are CEST
    * October is CEST up to the last sunday, 1am UTCu, then CET
    * A month with 31 days cannot have a last sunday earlier than the 25th.
    """
    month, day = utc_dt.month, utc_dt.day
    if month in [4, 5, 6, 7, 8, 9]:
        return CEST_OFFSET
    elif month in [1, 2, 11, 12]:
        return CET_OFFSET
    elif month == 3:
        if day < 25:
            return CET_OFFSET
        else:  # days 25 to 31
            return get_utc_offset_reference(utc_dt)
    else:  # month == 10
        if day < 25:
            return CEST_OFFSET
        else:  # days 25 to 31
            return get_utc_offset_reference(utc_dt)


def utc_to_cet_or_cest(utc_dt):
    """
    Returns a datetime in CET or CEST, whatever is used at that time.
    """
    utc_offset = datetime.timedelta(hours=get_utc_offset(utc_dt))
    return utc_dt + utc_offset



from unittest import TestCase

from zaach.math import prod


class ProdTests(TestCase):
    def test_prod_1_to_6(self):
        iterable = [1, 2, 3, 4, 5, 6]
        self.assertEqual(prod(iterable), 720)

    def test_prod_empty_iterable_raises_typeerror(self):
        with self.assertRaises(TypeError):
            prod([])

    def test_prod_non_digit_iterable_raises_typeerror(self):
        with self.assertRaises(TypeError):
            prod(["a", "b"])



class Borg(object):
    _shared_state = {}

    def __new__(cls, *args, **kwargs):
        obj = super(Borg, cls).__new__(cls, *args, **kwargs)
        obj.__dict__ = cls._shared_state
        return obj



import unittest

from zaach import strings


class StringsTestCase(unittest.TestCase):

    def test_lcut_1(self):
        s = '\'foo\''
        sub = '\''
        self.assertEqual(strings.lcut(s, sub), 'foo\'')

    def test_lcut_2(self):
        s = ''
        sub = '\''
        self.assertEqual(strings.lcut(s, sub), '')

    def test_rcut_1(self):
        s = '\'foo\''
        sub = '\''
        self.assertEqual(strings.rcut(s, sub), '\'foo')

    def test_rcut_2(self):
        s = ''
        sub = '\''
        self.assertEqual(strings.rcut(s, sub), '')

    def test_cut(self):
        s = '\'foo\''
        sub = '\''
        self.assertEqual(strings.cut(s, sub), 'foo')



import unittest

from zaach.oop import borg


class BorgTests(unittest.TestCase):

    def test_assimilation(self):
        drone = borg.Borg()
        drone.hello_world = "We are the Borg"

        locutus = borg.Borg()
        self.assertEqual(locutus.hello_world, "We are the Borg")


if __name__ == '__main__':
    unittest.main()



import random
import time
import unittest

from zaach.oop import cache


class CacheObject(cache.CacheMixin, object):
    pass


class CacheTests(unittest.TestCase):

    def test_cache_dict(self):
        co1 = CacheObject()
        co1cache = co1._cache
        self.assertIsInstance(co1cache, dict)

        co2 = CacheObject()
        co2cache = co2._cache
        self.assertIsInstance(co2cache, dict)

        self.assertNotEqual(id(co1cache), id(co2cache))

    def test_cacheget_and_cacheset_with_expiration(self):
        co = CacheObject()
        rnd = random.random()

        self.assertIsNone(co.cacheget('rnd'))
        co.cacheset('rnd', rnd, 0.01)
        self.assertEqual(co.cacheget('rnd'), rnd)
        self.assertEqual(co.cacheget('rnd'), rnd)

        time.sleep(0.01)
        self.assertIsNone(co.cacheget('rnd'))

    def test_cacheget_and_cacheset_without_expiration(self):
        co = CacheObject()
        rnd = random.random()

        self.assertIsNone(co.cacheget('rnd'))
        co.cacheset('rnd', rnd)
        self.assertEqual(co.cacheget('rnd'), rnd)
        self.assertEqual(co.cacheget('rnd'), rnd)


if __name__ == '__main__':
    unittest.main()



def lcut(s, sub):
    """
    Cuts off the substring `sub` if string `s` starts with it.
    """
    len_sub = len(sub)
    if s[:len_sub] == sub:
        return s[len_sub:]
    return s


def rcut(s, sub):
    """
    Cuts off the substring `sub` if string `s` ends with it.
    """
    len_sub = len(sub)
    if s[-len_sub:] == sub:
        return s[:-len_sub]
    return s


def cut(s, sub):
    """
    Cuts off the substring `sub` from the start and end of string `s`.
    """
    return rcut(lcut(s, sub), sub)



import datetime
import random
import unittest

from zaach.time import tzconversion


class TimezonConversionTests(unittest.TestCase):

    def test_last_sunday(self):
        for year in range(2011, 2100):
            for month in range(1, 13):
                sunday = tzconversion.get_last_sunday(year, month)
                self.assertEqual(sunday.weekday(), 6)

    def test_reference_against_performant_implementation(self):
        start_year = 1900 + int(random.random() * 300)
        for year in range(start_year, start_year + 2):
            for month in range(1, 13):
                for day in range(1, 29):
                    for hour in range(0, 24):
                        utc_dt = datetime.datetime(
                            year, month, day, hour,
                            int(random.random() * 60),
                            int(random.random() * 60)
                        )
                        offset_ref = tzconversion.get_utc_offset_reference(
                            utc_dt
                        )
                        offset = tzconversion.get_utc_offset(utc_dt)
                        self.assertEqual(offset_ref, offset)

    def test_utcnow_in_local_time(self):
        dt_utc = datetime.datetime.utcnow()
        dt_local = tzconversion.utc_to_cet_or_cest(dt_utc)
        self.assertNotEqual(dt_utc, dt_local)
        self.assertTrue(dt_local > dt_utc)
        self.assertTrue((dt_local - dt_utc).seconds in [3600, 7200])

    def test_reference_cet_to_cest_barriers(self):
        date_tuples_cet_to_cest = [
            (2020, 3, 29), (2019, 3, 31), (2018, 3, 25), (2017, 3, 26),
            (2016, 3, 27), (2015, 3, 29), (2014, 3, 30), (2013, 3, 31),
            (2012, 3, 25), (2011, 3, 27), (2010, 3, 28), (2009, 3, 29),
            (2008, 3, 30), (2007, 3, 25), (2006, 3, 26), (2005, 3, 27),
            (2004, 3, 28), (2003, 3, 30), (2002, 3, 31), (2001, 3, 25),
            (2000, 3, 26), (1999, 3, 28),
        ]
        for year, month, day in date_tuples_cet_to_cest:
            for hour in [0, 1, 2]:
                dt_utc = datetime.datetime(year, month, day, hour)
                dt_local = tzconversion.utc_to_cet_or_cest(dt_utc)
                delta_seconds = (dt_local - dt_utc).seconds
                if hour == 0:
                    self.assertEqual(delta_seconds, 3600)
                else:
                    self.assertEqual(delta_seconds, 7200)

    def test_reference_cest_to_cet_barriers(self):
        date_tuples_cest_to_cet = [
            (2020, 10, 25), (2019, 10, 27), (2018, 10, 28), (2017, 10, 29),
            (2016, 10, 30), (2015, 10, 25), (2014, 10, 26), (2013, 10, 27),
            (2012, 10, 28), (2011, 10, 30), (2010, 10, 31), (2009, 10, 25),
            (2008, 10, 26), (2007, 10, 28), (2006, 10, 29), (2005, 10, 30),
            (2004, 10, 31), (2003, 10, 26), (2002, 10, 27), (2001, 10, 28),
            (2000, 10, 29), (1999, 10, 31),
        ]
        for year, month, day in date_tuples_cest_to_cet:
            for hour in [0, 1, 2]:
                dt_utc = datetime.datetime(year, month, day, hour)
                dt_local = tzconversion.utc_to_cet_or_cest(dt_utc)
                delta_seconds = (dt_local - dt_utc).seconds
                if hour == 0:
                    self.assertEqual(delta_seconds, 7200)
                else:
                    self.assertEqual(delta_seconds, 3600)


if __name__ == '__main__':
    unittest.main()



from distutils.core import setup
from setuptools import find_packages


setup(
    name='zaach',
    version='1.0.8',
    packages=find_packages(),
    author="Fabian Topfstedt",
    url="https://bitbucket.org/fabian/zaach",
    license='The MIT Licence',
    package_data={'': ['LICENCE', 'README.md']},
    long_description='A collection of helpers',
    keywords = ['base64url', 'timezone', 'conversion'],
    classifiers = [],
)



import unittest

from zaach import base64url


class Base64UrlTests(unittest.TestCase):

    def test_encode_drops_padding(self):
        unencoded = 'a'
        encoded = base64url.encode(unencoded)
        expected = 'YQ'
        self.assertEqual(encoded, expected)

    def test_encode_underscore(self):
        unencoded = 'subjects?_d=1'
        encoded = base64url.encode(unencoded)
        expected = 'c3ViamVjdHM_X2Q9MQ'
        self.assertEqual(encoded, expected)

    def test_encode_minus(self):
        unencoded = ''.join([chr(i) for i in range(128)])
        encoded = base64url.encode(unencoded)
        expected = (
            'AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIjJCUmJygpKissLS4v'
            'MDEyMzQ1Njc4OTo7PD0-P0BBQkNERUZHSElKS0xNTk9QUVJTVFVWV1hZWltcXV5f'
            'YGFiY2RlZmdoaWprbG1ub3BxcnN0dXZ3eHl6e3x9fn8'
        )
        self.assertEqual(encoded, expected)

    def test_decode_adds_padding(self):
        encoded = 'YQ'
        unencoded = base64url.decode(encoded)
        expected = 'a'
        self.assertEqual(unencoded, expected)

    def test_decode_underscore(self):
        encoded = 'c3ViamVjdHM_X2Q9MQ'
        unencoded = base64url.decode(encoded)
        expected = 'subjects?_d=1'
        self.assertEqual(unencoded, expected)

    def test_decode_minus(self):
        encoded = (
            'AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIjJCUmJygpKissLS4v'
            'MDEyMzQ1Njc4OTo7PD0-P0BBQkNERUZHSElKS0xNTk9QUVJTVFVWV1hZWltcXV5f'
            'YGFiY2RlZmdoaWprbG1ub3BxcnN0dXZ3eHl6e3x9fn8'
        )
        unencoded = base64url.decode(encoded)
        expected = ''.join([chr(i) for i in range(128)])
        self.assertEqual(unencoded, expected)
